const {ccclass, property, menu, playOnFocus, executeInEditMode} = cc._decorator;

/** 
 * 状态渐变转换, 用法和 StateSwitch 基本相同
 */
@ccclass
@playOnFocus
@executeInEditMode
@menu('base/Flow')
export default class Flow extends cc.Component {

    @property(cc.Integer)
    _state = 0

    @property({type:cc.Integer})
    get state(){
        return this._state
    }

    set state(val){
        this._state = val
        this.subStates.forEach(s=> {
            s.getComponents(Flow).forEach(v=>v.state = this._state)
            s.getComponents("StateSwitch").forEach(v=>v.state = this._state)
        })
    }

    _lastState = -1
    
    @property(cc.Node)
    targetNode: cc.Node = null

    @property()
    propKey = ''

    @property([cc.Float])
    values: number[] = []

    @property
    time = 0

    /* 状态传递, 包含 Flow 或者 StateSwitch 的节点 */
    @property([cc.Node])
    subStates: Array<cc.Node> = []

    get val(){
        return this.getValue()
    }

    set val(val){
        this.setValue(val)
    }

    getValue(){
        let node = this.getTargetNode()
        if (this.propKey in node) {
            return node[this.propKey]
        }
        for (const comp of node.getComponents(cc.Component)) {
            if (this.propKey in comp) {
                return comp[this.propKey]
            }
        }
        return null
    }

    setValue(val){
        let node = this.getTargetNode()
        if (this.propKey in node) {
            node[this.propKey] = val
        }
        for (const comp of node.getComponents(cc.Component)) {
            if (this.propKey in comp) {
                comp[this.propKey] = val
            }
        }
    }

    getTargetNode(){
        return this.targetNode || this.node
    }

    update(dt){
        if (this._lastState != this.state && this.time) {
            let toVal = this.values[this.state]
            if (!cc.js.isNumber(toVal)) {
                return
            }
            let fromVal = this.values[this._lastState]
            if (!cc.js.isNumber(fromVal)) {
                this._lastState = this.state
                this.val = toVal
                return
            }
            let speed = (toVal - fromVal)/this.time
            this.val += speed * dt
            if (this.val >= toVal && toVal >= fromVal) {
                this._lastState = this.state
                this.val = toVal
            }else{
                if (this.val <= toVal && toVal < fromVal) {
                    this._lastState = this.state
                    this.val = toVal
                }
            }
        }else{
            if (cc.js.isNumber(this.values[this.state])) {
                this.val = this.values[this.state]
            }
        }
    }
}
