
const {ccclass, property, menu} = cc._decorator;

/**
 * 节点池对象
 */
@ccclass
@menu('base/Pool')
export default class Pool extends cc.Component {

    @property({
        tooltip: "模板对象是否也放入节点池, 放入节点池必须被其他组件引用, 否则会被回收"
    })
    putInPool = false

    /* 模板对象放入节点池之前保存其父节点 */
    parent:cc.Node = null

    idx = -1

    get mould(){
        return this.idx == -1 && !this.mouldItem
    }

    _nodePool = null

    /* 节点对象池 */
    get nodePool(){
        if (!this.mould) {
            return
        }
        if (this._nodePool == null) {
            this._nodePool = new cc.NodePool()
        }
        return this._nodePool
    }

    _items = {}

    _idPool = 0

    /* 模板对象 */
    mouldItem:Pool = null

    /* 自定义数据 */
    data:any = null

    onLoad(){
        if (this.mould) {
            this.parent = this.node.parent
            if (this.putInPool) {
                this.nodePool.put(this.node)
            }else{
                this.node.active = false
            }
        }else{
            if (this.mouldItem && this.idx == -1) {
                this.mouldItem._addItem(this)
            }
        }
    }

    _addItem(item){
        if (!this.mould) {
            return
        }
        item.idx = this._idPool
        this._idPool += 1
        item.mouldItem = this
        this._items[item.idx] = item.node
        this.node.emit('newItem', item.node)
    }

    getMouldItem():Pool{
        if (this.mould) {
            return this
        }
        return this.mouldItem
    }

    getItems():Array<cc.Node>;
    getItems<T extends cc.Component>(cls:{prototype: T}):Array<T>;
    getItems<T extends cc.Component>(cls?:{prototype: T}):Array<T|cc.Node>{
        let mould = this.getMouldItem()
        let items = []
        for (const key in mould._items) {
            if (mould._items.hasOwnProperty(key)) {
                const element = mould._items[key];
                if (cc.isValid(element)) {
                    if (cls) {
                        items.push(element.getComponent(cls))
                    }else{
                        items.push(element)
                    }
                }
            }
        }
        return items
    }

    getItem(idx):cc.Node{
        let mould = this.getMouldItem()
        let index = -1
        for (const key in mould._items) {
            if (mould._items.hasOwnProperty(key)) {
                const element = mould._items[key];
                if (cc.isValid(element)) {
                    index += 1
                    if (index == idx) {
                        return element
                    }
                }
            }
        }
    }

    findItem(func:(n:cc.Node)=>boolean):cc.Node{
        let mould = this.getMouldItem()
        for (const key in mould._items) {
            if (mould._items.hasOwnProperty(key)) {
                const element = mould._items[key];
                if (cc.isValid(element)) {
                    if (func(element)) {
                        return element
                    }
                }
            }
        }
    }

    findComp<T extends cc.Component>(cls:{prototype:T}, func:(n:T)=>{}):T{
        let mould = this.getMouldItem()
        for (const key in mould._items) {
            if (mould._items.hasOwnProperty(key)) {
                const element = mould._items[key];
                if (cc.isValid(element)) {
                    let comp = element.getComponent(cls)
                    if (comp && func(comp)) {
                        return comp
                    }
                }
            }
        }
    }

    newItem(parent?):cc.Node{
        let node = null
        let mouldItem = this.getMouldItem()
        if (!mouldItem) {
            return
        }
        let pool = mouldItem.nodePool
        if (pool.size() > 1) {
            node = pool.get()
        }else{
            node = cc.instantiate(mouldItem.node)
        }
        let item = node.getComponent('Pool')
        if (!parent) {
            parent = mouldItem.parent || mouldItem.node.parent
        }
        if (!parent) {
            cc.error("newItem no parent", this.node.name)
            return
        }
        mouldItem._addItem(item)
        node.parent = parent
        node.active = true
        return node
    }

    /* 由数据创建多个 */
    setDatas<T>(data:Array<T>, func:(item:cc.Node, d:T, idx:number)=>void, parent?){
        this.delAllItems()
        // let items = this.getItems(Pool)
        // items = items.sort((a,b)=> b.idx - a.idx)
        for (let i = 0; i < data.length; i++) {
            const element = data[i];
            // let item = items[i] ?items[i].node : this.newItem(parent)
            let item = this.newItem(parent)
            // cc.log('item', item.getComponent(Pool).idx)
            func(item, element, i)
        }
        // for (let i = data.length; i < items.length; i++) {
        //     let item = items[i]
        //     if (item) {
        //         item.delItem()
        //     }
        // }
    }

    setCount(count:number, func:(item:cc.Node, idx:number)=>void, parent?){
        this.delAllItems()
        for (let i = 0; i < count; i++) {
            let item = this.newItem(parent)
            func(item, i)
        }
    }

    addDatas(data:Array<any>, func:(item:cc.Node, d:any, idx:number)=>void, parent?){
        let count = this.getItemCount()
        for (let i = 0; i < data.length; i++) {
            const element = data[i];
            let item = this.newItem(parent)
            func(item, element, i+count)
        }
    }

    addCount(count:number, func:(item:cc.Node, idx:number)=>void, parent?){
        let last = this.getItemCount()
        for (let i = 0; i < count; i++) {
            let item = this.newItem(parent)
            func(item, i + last)
        }
    }

    getItemCount(){
        let count = 0
        let items = this.getMouldItem()._items
        for (const key in items) {
            if (items.hasOwnProperty(key)) {
                if (cc.isValid(items[key])) {
                    count += 1
                }
            }
        }
        return count
    }

    delItem(){
        if (!cc.isValid(this.node)) {
            return
        }
        if (this.mould) {
            return
        }
        let item = this.getMouldItem()
        if (!item) {
            return
        }
        if (!item._items) {
            return
        }
        delete item._items[this.idx]
        // this.node.active = false
        item.node.emit('delItem', this.node)
        this.node.stopAllActions()
        this.data = null
        item.nodePool.put(this.node)
        return true
    }

    delAllItems(){
        if (!cc.isValid(this.node)) {
            return
        }
        for (const item of this.getItems()) {
            item.getComponent('Pool').delItem()
        }
    }
}
