import Flow from "./Flow";
import { pLength } from "./fn";

const {ccclass, property, menu} = cc._decorator;

/** 
 * 属性
*/
@ccclass('Prop')
export class Prop {

    @property(cc.Node)
    node:cc.Node = null

    @property({
        tooltip: "组件名称, 为空取 node"
    })
    comp = ""

    @property({
        tooltip: "目标的属性名称"
    })
    prop = ""

    @property()
    _val = null

    @property()
    get val(){
        if (this._val !== null) {
            return this._val
        }
        let target = this.getTarget()
        if (target) {
            if (this.prop) {
                return target[this.prop]
            }
        }
        return null
    }

    set val(val){
        this._val = val
    }

    getTarget(){
        if (this.node) {
            if (this.comp) {
                return this.node.getComponent(this.comp)
            }
            return this.node
        }
    }
}


/**
 * !#zh 
 * @class StateSwitch
 */
@ccclass
@menu("base/StateSwitch")
export default class StateSwitch extends cc.Component {

    @property(cc.Integer)
    _state = 0

    @property({type:cc.Integer})
    get state(){
        return this._state
    }

    set state(val){
        this._state = val
        this.flush()
    }

    @property({
        type: [cc.String],
        tooltip: "状态描述",
    })
    stateDescs = []

    @property(cc.Node)
    targetNode:cc.Node = null

    @property([cc.SpriteFrame])
    spriteFrames = []

    @property([cc.Color])
    colors = []

    @property([cc.String])
    strings = []
    
    @property([cc.Node])
    nodeActives = []

    @property({
        type: cc.Float,
        tooltip: "旋转角度"
    })
    angles: number[] = []
    
    @property({
        type: cc.Float,
        tooltip: "透明度"
    })
    opacitys: number[] = []
    
    @property({
        type: cc.Float,
        tooltip: "缩放"
    })
    scales: number[] = []
    
    @property([cc.Size])
    sizes:Array<cc.Size> = []

    @property([cc.Vec2])
    positions:Array<cc.Vec2> = []

    @property(cc.Float)
    moveSpeed = 0
    
    @property({
        type: cc.Boolean,
        tooltip: "记录当前坐标到 positions 属性中"
    })
    get savePosition(){
        return false
    }
    set savePosition(val){
        if (CC_EDITOR) {
            this.positions[this.state] = cc.v2(this.node.x, this.node.y)
        }
    }
    
    /* 状态传递 */
    @property([StateSwitch])
    subStates: Array<StateSwitch> = []

    /* 状态传递 */
    @property([Flow])
    subFlows: Array<Flow> = []

    @property([Prop])
    _props: Prop[] = []

    @property({type:Prop})
    get props(){
        return this._props
    }

    set props(val){
        val.forEach(v=> {
            if (!v.node) {
                v.node = this.node
            }
        })
        this._props = val
    }

    flush (){
        let targetNode = this.targetNode || this.node
        let spriteFrame = this.spriteFrames[this.state]
        if (spriteFrame != undefined) {
            targetNode.getComponent(cc.Sprite).spriteFrame = spriteFrame
        }
        let color = this.colors[this.state]
        if (color) {
            targetNode.color = color
        }
        let str = this.strings[this.state]
        if (str != undefined) {
            targetNode.getComponent(cc.Label).string = str
        }
        this.nodeActives.forEach((node, idx)=>{
            if (node && this.state != idx) {
                node.active = false
            }
        })
        this.nodeActives.forEach((node, idx)=>{
            if (node && this.state == idx) {
                node.active = true
            }
        })
        let pos = this.positions[this.state]
        if (pos) {
            if (this.moveSpeed <= 0 || CC_EDITOR) {
                targetNode.position = pos
            }
        }
        
        let angle = this.angles[this.state]
        if (angle != undefined) {
            targetNode.angle = angle
        }
        let opacity = this.opacitys[this.state]
        if (opacity != undefined) {
            targetNode.opacity = opacity
        }
        let scale = this.scales[this.state]
        if (scale != undefined) {
            targetNode.scale = scale
        }
        let size = this.sizes[this.state]
        if (size != undefined) {
            targetNode.setContentSize(size)
        }
        let prop = this.props[this.state]
        if (prop != undefined) {
            let target = prop.getTarget()
            if (target) {
                if (prop.prop) {
                    target[prop.prop] = prop.val
                }
            }
        }
        this.getComponents(Flow).forEach(v=> {
            if (v.uuid != this.uuid) {
                v.state = this.state
            }
        })
        for (const s of this.subStates) {
            if (s) {
                s.getComponents(StateSwitch).forEach(v=> v.state = this.state)
            }
        }
        for (const s of this.subFlows) {
            if (s) {
                s.getComponents(Flow).forEach(v=> v.state = this.state)
            }
        }
        this.getComponents(Flow).forEach(v=> v.state = this.state)
    }

    update(dt){
        if (this.moveSpeed > 0) {
            let targetNode = this.targetNode || this.node
            let pos = cc.v2(targetNode.position)
            let to = this.positions[this.state]
            if(!to){return}
            let sub = to.sub(pos)
            if (pLength(sub) < this.moveSpeed/100) {
                targetNode.position = to
            }else{
                targetNode.position = pos.add(sub.normalize().mul(dt * this.moveSpeed))
            }
        }
    }

}
